const THREE = require('three');
const Trajectory = require('../server/model/trajectory');

///////////////////////////////////////////////////////////////////////////////
// The following classes are used to store the client model data.
// There is one instance of each class which is persistent:
// - Updated when processing the response of a request on the server.
// - Accessed by different views of the client.

/**
 * Mode.
 */
class Mode {
    constructor() {
        this._nomadConnection = false;
        this._nomadId = -1;
        this._spyMode = false;
    }

    set nomadConnection(value) { this._nomadConnection = value; }
    set nomadId(value) { this._nomadId = value; }
    set spyMode(value) { this._spyMode = value; } 

    get nomadConnection() { return this._nomadConnection; }
    get nomadId() { return this._nomadId; }
    get spyMode() { return this._spyMode; }

    update(modeModel) {
        this._nomadConnection = modeModel.nomadConnection;
        this._nomadId = modeModel.nomadId;
        this._spyMode = modeModel.spyMode;
    }
}

/**
 * ExpData data.
 */
class ExpData {
    constructor() {
        this._sqwtype = '';
        this._texformula = '';
        this._texwidth = 0;
        this._texheight = 0;
        this._texminsqw = 0;
        this._texmaxsqw = 0;
        this._texE = 0;
        this._texESigma = 0;
        this._filename = null;
        this._expSqwData = null;
    }

    get sqwtype() { return this._sqwtype; }
    get texformula() { return this._texformula; }
    get texwidth() { return this._texwidth; }
    get texheight() { return this._texheight; }
    get texminsqw() { return this._texminsqw; }
    get texmaxsqw() { return this._texmaxsqw; }
    get texE() { return this._texE; }
    get texESigma() { return this._texESigma; }
    get filename() { return this._filename; }
    get expSqwData() { return this._expSqwData; }
    set expSqwData(value) { this._expSqwData = value; }

    update(expDataModel) {
        this._sqwtype = expDataModel.sqwtype;
        this._texformula = expDataModel.texformula;
        this._texwidth = expDataModel.texwidth;
        this._texheight = expDataModel.texheight;
        this._texminsqw = expDataModel.texminsqw;
        this._texmaxsqw = expDataModel.texmaxsqw;
        this._texE = expDataModel.texE;
        this._texESigma = expDataModel.texESigma;
        this._filename = expDataModel.filename;

        if (expDataModel.expSqwData === undefined) {
            this._expSqwData = null;
        }
    }

}

/**
 * Grid data.
 */
class GridData {

    constructor() {
        this._numDiv = 1;
        this._delta = 0;
        this._adaptative = true;
    }
    
    get numDiv() { return this._numDiv; }
    set numDiv(value) { this._numDiv = value; }
    
    get delta() { return this._delta; }
    set delta(value) { this._delta = value; }

    get adaptative() { return this._adaptative; }
    set adaptative(value) { this._adaptative = value; }

    equals(gridModel) {
        if (this._numDiv !== gridModel.numDiv
        || this._delta !== gridModel.delta
        || this._adaptative !== gridModel.adaptative) {
            return false;
        }

        return true;
    }

    update(gridModel) {

        if (this.equals(gridModel)) {
            return false;
        }

        this._numDiv = gridModel.numDiv;
        this._delta = gridModel.delta;
        this._adaptative = gridModel.adaptative;

        return true;
    }
}

/**
 * Scattering data.
 */
class ScatteringData {

    constructor() {
        this._ki = 0;
        this._kf = 0;
        this._q = 0;
        this._qh = 0;
        this._qk = 0;
        this._ql = 0;

        this._kiPoints = {};
        this._kfPoints = {};

        this._a1 = 0;
        this._a2 = 0;
        this._a3 = 0;
        this._a3p = 0;
        this._a4 = 0;
        this._a5 = 0;
        this._a6 = 0;

        this._fixedKi = false;
        this._fixedKf = false;
        this._fixedQ = false;
        this._fixedDE = false;

        this._ss = 1;

        this._status = 0;
        this._paused = false;
    }
    
    get ki() { return this._ki; }
    set ki(value) { this._ki = value; }

    get kf() { return this._kf; }
    set kf(value) { this._kf = value; }

    get q() { return this._q; }
    set q(value) { this._q = value; }

    get qh() { return this._qh; }
    set qh(value) { this._qh = value; }

    get qk() { return this._qk; }
    set qk(value) { this._qk = value; }

    get ql() { return this._ql; }
    set ql(value) { this._ql = value; }

    get a1() { return this._a1; }
    set a1(value) { this._a1 = value; }

    get a2() { return this._a2; }
    set a2(value) { this._a2 = value; }

    get a3() { return this._a3; }
    set a3(value) { this._a3 = value; }

    get a3p() { return this._a3p; }
    set a3p(value) { this._a3p = value; }

    get a4() { return this._a4; }
    set a4(value) { this._a4 = value; }

    get a5() { return this._a5; }
    set a5(value) { this._a5 = value; }

    get a6() { return this._a6; }
    set a6(value) { this._a6 = value; }

    get kiPoints() { return this._kiPoints; }
    set kiPoints(value) { this._kiPoints = value; }

    get kfPoints() { return this._kfPoints; }
    set kfPoints(value) { this._kfPoints = value; }
    
    get fixedKi() { return this._fixedKi; }
    set fixedKi(value) { this._fixedKi = value; }

    get fixedKf() { return this._fixedKf; }
    set fixedKf(value) { this._fixedKf = value; }

    get fixedQ() { return this._fixedQ; }
    set fixedQ(value) { this._fixedQ = value; }

    get fixedDE() { return this._fixedDE; }
    set fixedDE(value) { this._fixedDE = value; }

    get ss()  { return this._ss; }
    set ss(value) { this._ss = value; }

    get status() { return this._status; }

    get paused() { return this._paused; }
    set paused(value) { this._paused = value; }

    update(scatteringModel) {

        if (this._paused) {
            return;
        }

        this._ki = scatteringModel.ki;
        this._kf = scatteringModel.kf;
        this._q = scatteringModel.q;
        this._qh = scatteringModel.qh;
        this._qk = scatteringModel.qk;
        this._ql = scatteringModel.ql;

        let kiPoints = scatteringModel.kiPoints;
        let kfPoints = scatteringModel.kfPoints;

        this._kiPoints = {
            start: new THREE.Vector3(kiPoints.start.x, kiPoints.start.y, kiPoints.start.z),
            end: new THREE.Vector3(kiPoints.end.x, kiPoints.end.y, kiPoints.end.z)
        },
        this._kfPoints = {
            start: new THREE.Vector3(kfPoints.start.x, kfPoints.start.y, kfPoints.start.z),
            end: new THREE.Vector3(kfPoints.end.x, kfPoints.end.y, kfPoints.end.z)
        },
        this._a1 = scatteringModel.a1;
        this._a2 = scatteringModel.a2;
        this._a3 = scatteringModel.a3;
        this._a3p = scatteringModel.a3p;
        this._a4 = scatteringModel.a4;
        this._a5 = scatteringModel.a5;
        this._a6 = scatteringModel.a6;

        this._fixedKi = scatteringModel.fixedKi;
        this._fixedKf = scatteringModel.fixedKf;
        this._fixedQ = scatteringModel.fixedQ;
        this._fixedDE = scatteringModel.fixedDE;

        this._ss = scatteringModel.ss;

        this._status = scatteringModel.status;
    }
}

/**
 * Plane data.
 */
class PlaneData {
    constructor() {
        this._ax = 1;
        this._ay = 0;
        this._az = 0;
        this._bx = 0;
        this._by = 1;
        this._bz = 0;

        this._aVector = new THREE.Vector3(1, 0, 0);
        this._bVector = new THREE.Vector3(0, 1, 0);
        this._abVector = new THREE.Vector3(0, 0, 1);
    }

    get ax() { return this._ax; }
    get ay() { return this._ay; }
    get az() { return this._az; }
    get bx() { return this._bx; }
    get by() { return this._by; }
    get bz() { return this._bz; }
    get aVector() { return this._aVector; }
    get bVector() { return this._bVector; }
    get abVector() { return this._abVector; }

    equals(planeModel) {

        if (this._ax !== planeModel.ax
        || this._ay !== planeModel.ay 
        || this._az !== planeModel.az
        || this._bx !== planeModel.bx
        || this._by !== planeModel.by
        || this._bz !== planeModel.bz) {
            return false;
        }

        return true;
    }

    update(planeModel) {

        if (this.equals(planeModel)) {
            return false;
        }

        this._ax = planeModel.ax;
        this._ay = planeModel.ay; 
        this._az = planeModel.az;
        this._bx = planeModel.bx; 
        this._by = planeModel.by; 
        this._bz = planeModel.bz;

        this._aVector = new THREE.Vector3(planeModel.aVector.x, planeModel.aVector.y, planeModel.aVector.z);
        this._bVector = new THREE.Vector3(planeModel.bVector.x, planeModel.bVector.y, planeModel.bVector.z);
        this._abVector = new THREE.Vector3(planeModel.abVector.x, planeModel.abVector.y, planeModel.abVector.z);

        return true;
    }
}

/**
 * Peaks data.
 */
class PeaksData {
    constructor() {
        this._lowX = 0;
        this._lowY = 0;
        this._lowZ = 0;
        this._upperX = 1;
        this._upperY = 1;
        this._upperZ = 1;
        this._sphereOpacity = 0.5;
    }

    get lowX() { return this._lowX; }
    get lowY() { return this._lowY; }
    get lowZ() { return this._lowZ; }
    get upperX() { return this._upperX; }
    get upperY() { return this._upperY; }
    get upperZ() { return this._upperZ; }
    get sphereOpacity() { return this._sphereOpacity; }

    equals(peaksModel) {
        if (this._lowX !== peaksModel.lowX
        || this._lowY !== peaksModel.lowY
        || this._lowZ !== peaksModel.lowZ
        || this._upperX !== peaksModel.upperX
        || this._upperY !== peaksModel.upperY
        || this._upperZ !== peaksModel.upperZ
        || this._sphereOpacity !== peaksModel.sphereOpacity) {
            return false;
        }

        return true;
    }

    update(peaksModel) {

        if (this.equals(peaksModel)) {
            return false;
        }

        this._lowX = peaksModel.lowX;
        this._lowY = peaksModel.lowY;
        this._lowZ = peaksModel.lowZ;
        this._upperX = peaksModel.upperX;
        this._upperY = peaksModel.upperY;
        this._upperZ = peaksModel.upperZ;
        this._sphereOpacity = peaksModel.sphereOpacity;

        return true;
    }
}

/**
 * Sample data.
 */
class SampleData {
    constructor() {
        this._as = 0;
        this._bs = 0;
        this._cs = 0;
        this._aa = 0;
        this._bb = 0;
        this._cc = 0;
        this._sg = 0;
        this._fixedValues = [];
        this._sync = true;
    }

    get as() { return this._as; }
    get bs() { return this._bs; }
    get cs() { return this._cs; }
    get aa() { return this._aa; }
    get bb() { return this._bb; }
    get cc() { return this._cc; }
    get sg() { return this._sg; }
    get fixedValues() { return this._fixedValues; }
    get sync() { return this._sync; }

    equals(sampleModel) {

        if (this._as !== sampleModel.as
        || this._bs !== sampleModel.bs
        || this._cs !== sampleModel.cs
        || this._aa !== sampleModel.aa
        || this._bb !== sampleModel.bb
        || this._cc !== sampleModel.cc
        || this._sg !== sampleModel.sg
        || this._sync !== sampleModel.sync) {
            return false;
        }

        return true;
    }
    
    update(sampleModel) {

        if (this.equals(sampleModel)) {
            return false;
        }

        this._as = sampleModel.as;
        this._bs = sampleModel.bs;
        this._cs = sampleModel.cs;
        this._aa = sampleModel.aa;
        this._bb = sampleModel.bb;
        this._cc = sampleModel.cc;
        this._sg = sampleModel.sg;
        this._fixedValues = sampleModel.fixedValues;
        this._sync = sampleModel.sync;

        return true;
    }
}

/**
 * Atom data.
 */
class AtomData {
    constructor(name, x, y, z) {
        this._name = name;
        this._x = x;
        this._y = y;
        this._z = z;
    }

    get name() { return this._name; }
    get x() { return this._x; }
    get y() { return this._y; }
    get z() { return this._z; }
}

/**
 * Atoms data.
 */
class AtomsData {
    constructor() {
        this._atoms = [];
        this._atomnames = [];
        this._atompos = [];
    }

    get atoms() { return this._atoms; }
    get atompos() { return this._atompos; }
    get atomnames() { return this._atomnames; }

    addAtom(name, x, y, z) {
        this._atoms.push(new AtomData(name, x, y, z));
    }

    setAtom(index, name, x, y, z) {
        this._atoms[index] = new AtomData(name, x, y, z);
    }

    updateAtomList() {
        this._atomnames = [];
        this._atompos = [];

        for(let atom of this._atoms) {
            this._atomnames.push(atom.name);
            this._atompos.push([ atom.x, atom.y, atom.z ]);
        }
    }

    clear() {
        this._atoms = [];
        this._atomnames = [];
        this._atompos = [];
    }

    equals(atomsModel) {

        let i = 0;
        for (let at of atomsModel.atoms) {

            if (i < this._atoms.length) {
                let atom = this._atoms[i];

                if (at.name !== atom.name 
                || at.x !== atom.x
                || at.y !== atom.y
                || at.z !== atom.z) {
                    //console.log('atom not equals');
                    return false;
                }
            }
            else {
                //console.log('size >');
                return false;
            }
            i++;
        }

        if (i < this._atoms.length) {
            //console.log('size <');
            return false;
        }

        //console.log('equals true');
        return true;
    }

    update(atomsModel) {

        if (this.equals(atomsModel)) {
            return false;
        }

        this._atoms = [];
        this._atomnames = [];
        this._atompos = [];

        for (let at of atomsModel.atoms) {
            this._atoms.push(new AtomData(at.name, at.x, at.y, at.z));
        }

        for (let name of atomsModel.atomnames) {
            this._atomnames.push(name);
        }

        for (let pos of atomsModel.atompos) {
            this._atompos.push(pos);
        }

        return true;
    }
}

/**
 * Matrix data.
 */
class MatrixData {

    constructor() {
        this._uMatrix = null;
        this._uMatrix_1 = null;
        this._bMatrix = null;
        this._bMatrix_1 = null;

        this._paused = false;
    }
    
    set uMatrix(value) { this._uMatrix = value.clone(); }
    get uMatrix() { return this._uMatrix; }

    set uMatrix_1(value) { this._uMatrix_1 = value.clone(); }
    get uMatrix_1() { return this._uMatrix_1; }
    
    set bMatrix(value) { this._bMatrix = value.clone(); }
    get bMatrix() { return this._bMatrix; }
    
    set bMatrix_1(value) { this._bMatrix_1 = value.clone(); }
    get bMatrix_1() { return this._bMatrix_1; }
    
    get paused() { return this._paused; }
    set paused(value) { this._paused = value; }

    update(matrixModel) {

        if (this._paused) {
            return;
        }

        let uMatrix = new THREE.Matrix3().fromArray(matrixModel.uMatrix);
        let bMatrix = new THREE.Matrix3().fromArray(matrixModel.bMatrix);

        if (this._uMatrix !== null && this._bMatrix !== null) {
            if (uMatrix.equals(this._uMatrix) && bMatrix.equals(this._bMatrix)) {
                return false;
            }
        }

        this._uMatrix = uMatrix;
        this._uMatrix_1 = new THREE.Matrix3().fromArray(matrixModel.uMatrix_1);
        this._bMatrix = bMatrix;
        this._bMatrix_1 = new THREE.Matrix3().fromArray(matrixModel.bMatrix_1);

        return true;
    }

}

/**
 * Trajectories data.
 */
class TrajectoriesData {

    constructor() {
        this._trajectories = {};
        this._visibility = {};
    }

    get trajectories() { return this._trajectories; }

    getTrajectoryById(id) { return this._trajectories[id]; }
    
    isVisible(id) {
        if (this._visibility[id] !== undefined) {
            return this._visibility[id];
        }
        return false;
    }

    setVisible(id, value) {
        this._visibility[id] = value;
    }

    addTrajectory(
        id,
        type,
        fixed,
        ki, kf, ss,
        startX, startY, startZ, startE, 
        endX, endY, endZ, endE,
        steps, mn, time, useTime, name, 
        uMatrix, bMatrix,
        startingTime) {
        
        this._trajectories[id] = new Trajectory(
                                        type,
                                        fixed,
                                        ki, kf, ss,
                                        new THREE.Vector4(startX, startY, startZ, startE), new THREE.Vector4(endX, endY, endZ, endE),
                                        steps, mn, time, useTime,
                                        uMatrix, bMatrix, 
                                        startingTime);

        this._trajectories[id]._id = id;
        this._trajectories[id]._name = name;
    }

    deleteTrajectory(id) {
        delete this._visibility[id];
    }

    deleteTrajectories() {
        this._visibility = {};
    }

    update(trajectoriesModel) {

        // Check size.
        let size = 0;
        for (let tid in trajectoriesModel) {
            size++;
        }

        for (let uid in this._trajectories) {
            size--;
        }

        // If size is equals, then the trajectories do not need to be updated.
        if (size === 0) {
            return false;
        }

        // Clear the trajectories.
        this._trajectories = {};

        // Add the trajectories.
        for (let id in trajectoriesModel) {
            let traj = trajectoriesModel[id];
            this.addTrajectory(id, 
                        traj.type,
                        traj.fixed,
                        traj.ki, traj.kf, traj.ss,
                        traj.startX, traj.startY, traj.startZ, traj.startE, 
                        traj.endX, traj.endY, traj.endZ, traj.endE, 
                        traj.steps, traj.mn, traj.time, traj.useTime, 
                        traj.name,
                        new THREE.Matrix3().fromArray(traj.uMatrix),
                        new THREE.Matrix3().fromArray(traj.bMatrix),
                        traj.startingTime);
        }

        return true;        
    }
}

/**
 * Trajectories data data.
 */
class TrajectoriesDataData {

    constructor() {
        this._data = {};
        this._currentId = 0;
    }
    
    get currentId() { return this._currentId; }
    get data() { return this._data; }

    get currentData() { 
        if (this._currentId !== 0) {
            return this._data[this._currentId];
        }
        return null;
    }

    update(trajectoriesDataModel) {
    
        this._currentId = trajectoriesDataModel.id;
        this._data = trajectoriesDataModel.data;

        return true;        
    }
}

///////////////////////////////////////////////////////////////////////////////
// Shared objects.
const mode = new Mode();

const expdata = new ExpData();

const grid = new GridData();

const plane = new PlaneData();

const peaks = new PeaksData();

const sample = new SampleData();

const atoms = new AtomsData();

const scattering = new ScatteringData();

const trajectory = new TrajectoriesData();

const trajectoryData = new TrajectoriesDataData();

const matrix = new MatrixData();

const config = {};

module.exports = { mode, expdata, grid, plane, peaks, sample, atoms, scattering, trajectory, trajectoryData, matrix, config };