const Controller = require('./controller');
const PubSub = require('pubsub-js');
const { requester } = require('../main/context');
const { atoms, sample } = require('../app-state');

class UnitCellController extends Controller {
    constructor(atomsPanel) {
        super();

        this._atomsPanel = atomsPanel;

        // Actions handlers.
        PubSub.subscribe('add-atom', (msg, value) => { this.addAtom(value); });
        PubSub.subscribe('del-atom', (msg, value) => { this.delAtom(value.index, value.row); });
        PubSub.subscribe('clear-atoms', () => { this.clearAtoms(); });
        PubSub.subscribe('update-atom', (msg, value) => { this.updateAtom(value.index, value.newAtom); });
        PubSub.subscribe('update-atom-controller', () => { this._notifyViews(); })

        // Event handlers.
        PubSub.subscribe('change-atoms-conf', () => { this.atomsChange(); });

        // The atoms must be added in case of a new client with the server having already some atoms from a previous session.
        this.init();
    }

    init() {
        this._atomsPanel.clearAtoms();

        for(let at of atoms.atoms) {
            this._atomsPanel.addAtom(at.name, at.x, at.y, at.z);
        }

        PubSub.publish('redraw-unitcell', { atomnames: atoms.atomnames, atompos: atoms.atompos });
    }

    /**
     * Update atoms in response to model update.
     */
    atomsChange() {
        this._atomsPanel.clearAtoms();

        for(let at of atoms.atoms) {
            this._atomsPanel.addAtom(at.name, at.x, at.y, at.z);
        }
        this._notifyViews();
    }

    /**
     * Add an atom.
     */
    addAtom(atom) {
        requester.request('addAtom', atom, () => {
            atoms.addAtom(atom.name, atom.x, atom.y, atom.z);
            this._atomsPanel.addAtom(atom.name, atom.x, atom.y, atom.z);
            this._notifyViews();
        });
    }

    /**
     * Delete an atom. 
     */
    delAtom(index, atom) {
        requester.request('delAtom', index, () => {
            this._atomsPanel.delAtom(atom);
            atoms.atoms.splice(index, 1);
            this._notifyViews();
        });
    }

    /**
     * Clear all atoms.
     */
    clearAtoms() {
        requester.request('clearAtoms', 0, () => {
            atoms.clear();
            this._atomsPanel.clearAtoms();
            this._notifyViews();
        });
    }

    /**
     * Update an atom parameters
     */
    updateAtom(index, newAtom) {
        requester.request('updateAtom', { index: index, newAtom: newAtom }, () => {
            atoms.setAtom(index, newAtom.name, newAtom.x, newAtom.y, newAtom.z);
            this._notifyViews()
        });
    }

    /**
     * Notify view of atoms change.
     */
    _notifyViews() {
        this._updateAtomList();
        PubSub.publish('redraw-unitcell', { atomnames: atoms.atomnames, atompos: atoms.atompos });
        PubSub.publish('change-atoms', 0);
    }

    /**
     * Update the local atom list.
     */
    _updateAtomList() {
        atoms.updateAtomList();
    }

    /**
     * Get atom index in local list.
     */
    _getAtomIndex(name, x, y, z) {
        for(let i = 0; i < atoms.atoms.length; i++) {
            if(atoms.atoms[i].name == name && atoms.atoms[i].x == x && atoms.atoms[i].y === y && atoms.atoms[i].z === z) {
                return i;
            }
        }
    }
}

module.exports = UnitCellController;