const TrajectoryLine = require('./trajectory-line');
const LineSegment = require('./line-segment');
const {scale, vector3WorldToView, scalarWorldToView, vector3ViewToWorld} = require('../conversion');
const Color = require('../colors');
const {mapColor} = require('../map-color');
const { mode } = require('../../app-state');
const THREE = require('three');

class EnergyLine extends TrajectoryLine {
    constructor(type, parent, start, end, center, steps) {
        super(type, parent, start, end, center, steps);

        this._data = [];

        // The line is a segment.
        this._line = null;

        // Create a group for the data.
        this._dataGroup = new THREE.Group();
        this._group.add(this._dataGroup);

        this._deltaSphereGroup = new THREE.Group();
        this._dataGroup.add(this._deltaSphereGroup);

        this.highlight(false);
        this.updateLine();
    }

    setData(data) {
        this._data = data;
    }

    updateDataGroup() {
        
        // Reset the group transform.
        let inv = new THREE.Matrix4();
        inv.getInverse(this._dataGroup.matrix);
        this._dataGroup.applyMatrix(inv);

        // Then we can apply the new matrix.
        let matrix4 = this.calculateSegmentMatrix(this._start, this._end);
        this._dataGroup.applyMatrix(matrix4);
        
        // Remove old spheres.
        for (let i = this._deltaSphereGroup.children.length - 1; i >= 0; i--) {
            this._deltaSphereGroup.remove(this._deltaSphereGroup.children[i]);
        }
        
        // Calculate length of the segment.
        let length = this._end.clone().sub(this._start).length();
        
        // Calculate delta between spheres.
        let delta = length / this._intervals; 

        // Place new spheres.
        let dataGeometry = new THREE.SphereGeometry(0.5 * delta, 8, 8);
        let pointGeometry = new THREE.SphereGeometry(1 / scale, 8, 8);
        
        // Iterate the points and draw data sphere or point sphere.
        for (let i = 0; i < this._intervals + 1; i++) {

            let position = new THREE.Vector3(0, 0, i * delta);

            let sphere;

            // Draw a data sphere.
            if (i < this._data.length) {
                let dataMaterial = new THREE.MeshBasicMaterial({color: 0xffffff, depthTest: true, transparent: false});
                const color = mapColor(this._data[i], 0, 500);
                dataMaterial.color.setRGB(color.r, color.g, color.b);
                sphere = new THREE.Mesh(dataGeometry, dataMaterial);
            }
            // Draw a point sphere.
            else {
                let pointMaterial = new THREE.MeshBasicMaterial({color: 0xffffff, depthTest: true, transparent: false});
                sphere = new THREE.Mesh(pointGeometry, pointMaterial);
            }
            
            sphere.renderOrder = 4;
            sphere.position.copy(position);
            this._deltaSphereGroup.add(sphere);
        }
    }

    /**
     * Update line position and delta spheres.
     */
    updateLine() {

        // Create the line.
        let lineMaterial = new THREE.LineBasicMaterial({ color: Color.Q, linewidth: 5, depthTest: true, transparent: false});
        lineMaterial.color.set(this.getColor(true));

        // Remove the line from the data group.
        if (this._line !== null) {
            this._dataGroup.remove(this._line.object3D);
        }

        // Calculate length of the segment.
        let length = this._end.clone().sub(this._start).length();

        // Create a line segment for the line.
        let segment = new LineSegment(new THREE.Vector3(0, 0, 0), new THREE.Vector3(0, 0, length), 0.5/scale, lineMaterial);
        this._line = segment;

        // Add the line to the data group.
        this._dataGroup.add(this._line.object3D);

        // Update the data.
        this.updateDataGroup();
    }

    highlightLines(color) {
    }

    highlightSpheres(value) {
        
        // The spheres are always visible in spy mode.
        if (!mode.spyMode) {
            this._deltaSphereGroup.visible = value;
        }
        else {
            this._deltaSphereGroup.visible = true;
        }
    }

    /**
     * Set line visibility.
     */
    setVisibility(visibility) {
        this._line.object3D.visible = visibility;
        this._dataGroup.visible = visibility;
    }

    /**
     * Remove this trajectory properly.
     */
    destroy() {
        this._line.destroy();

        for (let d of this._deltaSphereGroup.children) {
            d.geometry.dispose();
            d.material.dispose();
        }
    }
}

module.exports = EnergyLine;