const TrajectoryLine = require('./trajectory-line');
const LineSegment = require('./line-segment');
const {scale, vector3WorldToView, scalarWorldToView, vector3ViewToWorld} = require('../conversion');
const Color = require('../colors');
const TrajectorySegmentIterator = require('./trajectory-segment-iterator');
const TrajectoryCircleIterator = require('./trajectory-circle-iterator');
const THREE = require('three');

class HKLLine extends TrajectoryLine {
    constructor(type, parent, start, end, center, steps) {
        super(type, parent, start, end, center, steps);

        let geometry = new THREE.SphereGeometry(1/scale, 8, 8);
        let material = new THREE.MeshBasicMaterial({color: 0x0000ff, opacity: 0.5, depthTest: true, transparent: false});

        // Create the trajectory start sphere
        this._startSphere = new THREE.Mesh(geometry, material);
        this._startSphere.renderOrder = 1;

        // Create the trajectory end sphere.
        this._endSphere = new THREE.Mesh(geometry, material);
        this._endSphere.renderOrder = 1;

        // The line is an array of segments.
        this._line = [];

        this._lineGroup = new THREE.Group();
        this._group.add(this._lineGroup);

        this._deltaSphereGroup = new THREE.Group();
        this._group.add(this._startSphere);
        this._group.add(this._endSphere);
        this._group.add(this._deltaSphereGroup);

        this.highlight(false);
        this.updateLine();
    }

    /**
     * Update line position and delta spheres.
     */
    updateLine() {

        // First copy the start and end position if they changed.
        this._startSphere.position.copy(this._start.clone());
        this._endSphere.position.copy(this._end.clone());
        
        let geometry = new THREE.SphereGeometry(1 / scale, 8, 8);
        let material = new THREE.MeshBasicMaterial({color: 0xffffff, depthTest: true, transparent: false});

        let iterator = null;
        if (this._type == 'q') {
            iterator = new TrajectorySegmentIterator(this._start, this._end, this._intervals);
        }
        else if (this._type == 'ki' || this._type == 'kf') {
            iterator = new TrajectoryCircleIterator(this._center, this._start, this._end, this._intervals);
        }

        // Empty the line.
        this._line = [];

        // Remove old line segments.
        for (let i = this._lineGroup.children.length - 1; i >= 0; i--) {
            this._lineGroup.remove(this._lineGroup.children[i]);
        }

        // Remove old spheres.
        for (let i = this._deltaSphereGroup.children.length - 1; i >= 0; i--) {
            this._deltaSphereGroup.remove(this._deltaSphereGroup.children[i]);
        }

        // Create the line.
        let lineMaterial = new THREE.LineBasicMaterial({ color: Color.Q, linewidth: 5, depthTest: true, transparent: false});
        lineMaterial.color.set(this.getColor(true));

        // Define the previous point for the line segments.
        let previousPoint = null;

        // Place new spheres and segments.
        for (let i = 0; i < this._intervals; i++) {

            // Get the current point.
            let currentPoint = iterator.value().clone();

            // Add the segment when the previous point exists.
            if (previousPoint !== null) {
                let length = currentPoint.clone().sub(previousPoint).length();
                // Add the segment if start and end points are different (bad drawing otherwise).
                if (length !== 0) {
                    let segment = new LineSegment(previousPoint, currentPoint, 0.5 / scale, lineMaterial);
                    this._line.push(segment);
                    this._lineGroup.add(segment.object3D);
                }
            }

            // Define a sphere.
            let stepSphere = new THREE.Mesh(geometry, material);
            stepSphere.renderOrder = 3;
            stepSphere.position.copy(iterator.value());
            this._deltaSphereGroup.add(stepSphere);

            // Set the previous point.
            previousPoint = currentPoint;

            // Next iteration.
            iterator.next();
        }
        
        // Last segment.
        if (previousPoint !== null) {
            let lastPoint = iterator.value().clone();
            let length = lastPoint.clone().sub(previousPoint).length();
            // Add the segment if start and end points are different (bad drawing otherwise).
            if (length !== 0) {
                let segment = new LineSegment(previousPoint, lastPoint, 0.5 / scale, lineMaterial);
                this._lineGroup.add(segment.object3D);
            }
        }
    }

    highlightLines(color) {
        
        // Change the color of the segments.
        for (let i = 0; i < this._line.length; i++) {
            this._line[i].material.color.set(color);
        }
    }

    highlightSpheres(value) {
        this._deltaSphereGroup.visible = value;
    }

    /**
     * Set line visibility.
     */
    setVisibility(visibility) {
        this._startSphere.visible = visibility;
        this._endSphere.visible = visibility;
        this._lineGroup.visible = visibility;
        this._deltaSphereGroup.visible = visibility;
    }

    /**
     * Remove this trajectory properly.
     */
    destroy() {
        this._startSphere.geometry.dispose();
        this._endSphere.geometry.dispose();
        this._startSphere.material.dispose();
        this._endSphere.material.dispose();

        for (let d of this._deltaSphereGroup.children) {
            d.geometry.dispose();
            d.material.dispose();
        }
    }
}

module.exports = HKLLine;