/**
 * Reciprocal 3D view.
 */

const myLayout = require('../../main/multi-layout');
const PubSub = require('pubsub-js');
const THREE = require('three');
const OrbitControls = require('three-orbitcontrols');
const DragControls = require('three-dragcontrols');
const { menuEvent, displayMode, receiver } = require('../../main/context');
const { mode, matrix } = require('../../app-state');

const viewManager = require('../../utils/view-manager');
const MapControls = require('./map-control');
const CameraMove = require('../../utils/camera-move');
const DeviceOrientationControls = require('../reciprocal/three-device-orientation');

const PlaneController = require('../../controllers/plane-controller');
const LatticeController = require('../../controllers/lattice-controller');
const PeaksController = require('../../controllers/peaks-controller');
const SampleController = require('../../controllers/sample-controller');
const ScatteringController = require('../../controllers/scattering-controller');
const TrajectoryController = require('../../controllers/trajectory-controller');
const UnitCellController = require('../../controllers/unitcell-controller');
//const ConfigurationController = require('../../controllers/configuration-controller');

const LatticeView3D = require('../3D/lattice-view');
const PlaneView3D = require('../3D/plane-view');
const ScatteringView3D = require('../3D/scattering-view');
const TrajectoryView3D = require('../3D/trajectory-view');

const PlanePanelView = require('../panels/plane-view');
const GridPanelView = require('../panels/grid-view');
const TrajectorySettingsPanelView = require('../panels/trajectory-settings-view');
const ExpDataPanelView = require('../panels/expdata-view');
const PeaksPanelView = require('../panels/peaks-view');
const SamplePanelView = require('../panels/sample-view');
const ScatteringPanelView = require('../panels/scattering-view');
const TrajectoryPanelView = require('../panels/trajectory-view');
const AnglesPanelView = require('../panels/angles-view');
const AtomsPanelView = require('../panels/atoms-view');
//const ConfigurationPanelView = require('../panels/configuration-view');

const Constants = require('../../utils/constants');

class ReciprocalView3D {

    constructor() {
        
        this._renderer = null;
        this._width = 100;
        this._height = 100;

        let html = `
        <div style='padding-right:0;padding-left:0;width:100%;height:100%;' id='reciprocalView3D-layout'/>
        <div class='p-1 position-absolute bg-secondary rounded text-white' id='coordinates' style='pointer-events: none; opacity: 0.7; font-size: 8pt'/>
        <div class='p-1 position-absolute bg-secondary rounded text-white' id='latticeCoord' style='pointer-events: none; opacity: 0.7; font-size: 8pt'/>
        `;

        // Using lambda function does not work here.
        myLayout.registerComponent('Reciprocal View', function(container, state) {
            container.getElement().html(html);
        });

        // But it works here.
        myLayout.on('componentCreated', (component) => {

            // We can get the component main at its creation.
            if (component.componentName === 'Reciprocal View') {

                // Store the initial size of the component.
                this._width = component.container.width;
                this._height = component.container.height;

                component.container.on('resize', () => {
                    this.onResize(component.container.width, component.container.height);
                });
            }
        });

        // Create the panels.
        this._anglesPanel = new AnglesPanelView();
        this._peaksPanel = new PeaksPanelView();
        this._scatteringPanel = new ScatteringPanelView();
        this._planePanel = new PlanePanelView();
        this._gridPanel = new GridPanelView();
        this._trajectorySettingsPanel = new TrajectorySettingsPanelView();
        this._samplePanel = new SamplePanelView();
        this._expDataPanel = new ExpDataPanelView();
        this._atomsPanel = new AtomsPanelView();
        //this._configPanel = new ConfigurationPanelView();
        this._trajectoryPanel = new TrajectoryPanelView();


        // The Reciprocal view must be added after the atoms panel which must be initialised before.
        viewManager.addView(this);
    }

    get renderer() { return this._renderer; }

    init() {
        // Selects the div to place the three.js scene.
        this._container = document.getElementById('reciprocalView3D-layout');

        // Antialias to true slows down the display but is required in the case of macOS.
        // Otherwise the display is flickering when there are multiple scissors.
        let antialias = false;
        if (process.platform === 'darwin') {
            antialias = true;
        }

        // Create the renderer.
        this._renderer = new THREE.WebGLRenderer({ antialias: antialias });

        // Set the initial size.
        this._renderer.setSize(this._width, this._height);
        this._container.appendChild(this._renderer.domElement);


        // Mouse vector.
        this._mouse = new THREE.Vector2();

        // Container of the kiSphere and qSphere: array of objects has to be passed to the drag controls in order to allow dragging.
        this._objects = [];

        // Variable to display only peaks in the scattering plane.
        this._displayOnlyPeaksInPlane = false;

        // Variable to display scattering plane in 2D.
        this._display2DView = false;
        this._controlToken = null;

        //////// DEFINE SCENE ///////////
        // Raycaster and renderer objects.
        this._raycaster = new THREE.Raycaster();

        // New scene.
        this._scene = new THREE.Scene();
        this._scene.userData.element = this._container;

        // Setup of camera.
        let viewAngle = 45, cameraAspect = this._width / this._height, cameraNear = 0.01, cameraFar = 1000;
        this._scene.userData.camera = new THREE.PerspectiveCamera(viewAngle, cameraAspect, cameraNear, cameraFar);

        //this._scene.userData.camera = new THREE.OrthographicCamera(-window.innerWidth/5, window.innerWidth/5, -window.innerHeight/5, window.innerHeight/5, cameraNear, cameraFar);
        this._scene.userData.camera.position.set(0, 0, 300);

        // Set the background according to the calculation status.
        //this.setBackground(scattering.status);
        this._scene.background = new THREE.Color(0xcccccc);

        // Add light to the scene.
        this._scene.add(new THREE.AmbientLight(0xaaaaaa));

        // Light source is always at the camera position.
        this._cameraLight = new THREE.PointLight(0xaaaaaa);
        this._cameraLight.position.copy(this._scene.userData.camera.position);
        this._scene.add(this._cameraLight);
        /////////////////////////////////

        //////// DEFINE CONTROLS //////////
        // Prepare Orbit controls. Enable zoom, rotation and pan.
        this._scene.userData.controls = new OrbitControls(this._scene.userData.camera, this._container);
        this._scene.userData.controls.enableZoom = true;
        this._scene.userData.controls.enableRotate = true;
        this._scene.userData.controls.enablePan = true;

        // Disable the arrow keys that conflict with the cursor movement in a text field.
        this._scene.userData.controls.keys = {
            LEFT: 0,
            UP: 0,
            RIGHT: 0,
            BOTTOM: 0
        }

        // Create the drag controls for the ki and kf spheres.
        this._dragControls = new DragControls(this._objects, this._scene.userData.camera, this._container);
        this._cameraMove = new CameraMove(this._scene.userData.camera);

        // Initialize MapControls here, otherwise event listeners are removed.
        this._scene.userData.controls2D = new MapControls(this._scene.userData.camera, this._container);

        // Disable 2D Controls
        this._scene.userData.controls2D.enabled = false;
        this._scene.userData.controls2D.enablePan = false;
        this._scene.userData.controls2D.enableZoom = false;

        //////// DEFINE EVENTS //////////
        this._container.addEventListener('mousemove', this.onDocumentMouseMove.bind(this), false);

        // Background color change if a position is forbidden.
        receiver.on('update-status', (msg, value) => {
            //this.setBackground(value.status);
        });

        // Registers callbacks.
        this.registerControls();
        this.registerActions();
        /////////////////////////////////

        //////// DEFINE MVC //////////
        // 3D views definition.
        let lattice3D = new LatticeView3D(this._scene, this._raycaster, this._renderer);
        let plane3D = new PlaneView3D(this._scene, this._raycaster, this._scene.userData.controls2D);
        let scattering3D = new ScatteringView3D(this._scene, this._dragControls, this._raycaster, this._objects, this._scene.userData.controls2D);
        let trajectory3D = new TrajectoryView3D(this._scene, this._dragControls, scattering3D);

        // Initialize controllers.
        this._planeController = new PlaneController(this._planePanel, this._gridPanel, this._expDataPanel, plane3D);
        this._latticeController = new LatticeController(lattice3D);
        this._peaksController = new PeaksController(this._peaksPanel);
        this._scatteringController = new ScatteringController(this._scatteringPanel, this._anglesPanel, scattering3D);
        this._trajectoryController = new TrajectoryController(this._trajectoryPanel, trajectory3D);
        this._unitcellController = new UnitCellController(this._atomsPanel);
        // The sample controller is initialised after the unit cell because in case of a new client 
        // with the server having already a space group configured from a previous session, the atoms must be initialised before the space group.
        this._sampleController = new SampleController(this._samplePanel);
        //this._configController = new ConfigurationController(this._configPanel);
        /////////////////////////////////

        let self = this;

        function setOrientationControls(e) {
            // If device orientation is not available, return.
            if (!e.alpha) {
                return;
            }

            // Create controls for mobile.
            self._scene.userData.controls = new DeviceOrientationControls(self._scene.userData.camera, true);
            self._scene.userData.controls.connect();
            self._scene.userData.controls.update();


            //$('body').on('touchstart', fullscreen);
            window.removeEventListener('deviceorientation', setOrientationControls, true);
        }

        window.addEventListener('deviceorientation', setOrientationControls, true);

        // Toggle mobile mode.
        if (/Android|webOS|iPhone|iPad|BlackBerry|Windows Phone|Opera Mini|IEMobile|Mobile/i.test(navigator.userAgent)) {
            menuEvent.dispatch('toggle-vr-view', true);
        }

        PubSub.subscribe('reset-reciprocal-view-3d', () => {
            this.reset();
        });

        // Start the spy loop.
        if (mode.spyMode) {
            this._trajectoryController.startSpy();
        }
        else {
            this._trajectoryController.stopSpy();
        }
    }

    setBackground(status) {
        this._scene.background = (status === 1) ? new THREE.Color(0xcccccc) : new THREE.Color(0xda9797);
    }

    reset() {
        this._planeController.init();
        this._latticeController.init();
        this._peaksController.init();
        this._sampleController.init();
        this._scatteringController.init();
        this._trajectoryController.init();
        this._unitcellController.init();
    }

    /**
     * Set app to full screen mode.
     */
    fullscreen() {
        let doc = document.body;
        if (doc.requestFullscreen) {
            doc.requestFullscreen();
        } else if (doc.msRequestFullscreen) {
            doc.msRequestFullscreen();
        } else if (doc.mozRequestFullScreen) {
            doc.mozRequestFullScreen();
        } else if (doc.webkitRequestFullscreen) {
            doc.webkitRequestFullscreen();
        }
    }

    /**
     * Register callbacks for menu actions.
     */
    registerActions() {
        menuEvent.on('toggle-peaks-in-plane', (event, value) => {
            this._displayOnlyPeaksInPlane = value;
            this._planeController.drawGrid();
            this._latticeController.setVisibility(value);
        });

        menuEvent.on('toggle-vr-view', (event, value) => {
            displayMode.vr = value;
            this.fullscreen();
        }); 
        
        menuEvent.on('toggle-2d-view', (event, value) => {
            // Hide/show plane.
            this._display2DView = value;
            this._displayOnlyPeaksInPlane = value;
            this._planeController.drawGrid();
            this._latticeController.setVisibility(value);

            if (value) {
                // Define start point and end point of the camera movement.
                let cameraPosition = new THREE.Vector3(0, 0, 300).applyMatrix3(matrix.uMatrix_1);
                this._cameraMove.clear();
                this._cameraMove.addWaypoint(this._scene.userData.camera.position, this._scene.userData.controls.target);
                this._cameraMove.addWaypoint(cameraPosition, new THREE.Vector3());

                // Disable OrbitControls
                this._scene.userData.controls.enabled = false;
                this._scene.userData.controls.enableRotate = false;
                this._scene.userData.controls.enablePan = false;
                this._scene.userData.controls.enableZoom = false;

                // Enable 2D Controls
                this._scene.userData.controls2D.enabled = true;
                this._scene.userData.controls2D.enablePan = true;
                this._scene.userData.controls2D.enableZoom = true;

                // Start movement.
                this._cameraMove.enableMove = true;

                // Update camera position when plane change.
                this._controlToken = PubSub.subscribe('change-plane', () => {
                    cameraPosition = new THREE.Vector3(0, 0, 300).applyMatrix3(matrix.uMatrix_1);
                    this._scene.userData.camera.position.copy(cameraPosition);
                    this._scene.userData.camera.lookAt(0, 0, 0);
                    this._scene.userData.controls2D.update();
                });
            } else {
                // Enable OrbitControls
                this._scene.userData.controls.enabled = true;
                this._scene.userData.controls.enableRotate = true;
                this._scene.userData.controls.enablePan = true;
                this._scene.userData.controls.enableZoom = true;

                // Disable 2D Controls
                this._scene.userData.controls2D.enabled = false;
                this._scene.userData.controls2D.enablePan = false;
                this._scene.userData.controls2D.enableZoom = false;

                this._scene.userData.controls.update();
                this._scene.userData.controls2D.update();

                PubSub.unsubscribe(this._controlToken);
            }
        });

        menuEvent.on('toggle-energy-view', (event, value) => {
            if (value) {
                displayMode.view = Constants.DisplayMode.SPE;
            }
            else {
                displayMode.view = Constants.DisplayMode.HKL;
            }

            // Update the lattice, only peaks in plane.
            this._latticeController.setVisibility(value);

            // Update the trajectories.
            this._trajectoryController.changeDisplayMode();
        });
    }

    /**
     * Register callbacks on controls.
     */
    registerControls() {
        this._scene.userData.controls.addEventListener('change', this.lightUpdate.bind(this));
        this._scene.userData.controls2D.on('change', this.lightUpdate.bind(this));

        // Disable controls when object is being dragged. Safety purposes.
        this._dragControls.addEventListener('dragstart', (event) => {
            this._scene.userData.controls.enabled = false;
            this._scene.userData.controls2D.enabled = false;
        });

        // Enable controls after drag has finished.
        this._dragControls.addEventListener('dragend', (event) => {
            this._scene.userData.controls.enabled = true;
            this._scene.userData.controls2D.enabled = true;
        });

        // Enable 2D controls when camera transition has finished.
        this._cameraMove.on('moveend', () => {
            this.lightUpdate();
            this._scene.userData.controls2D.update();
            this._scene.userData.controls2D.enabled = this._display2DView;
            this._scene.userData.controls2D.enablePan = this._display2DView;
            this._scene.userData.controls2D.enableZoom = this._display2DView;
        });
    }

    // getters
    get displayOnlyPeaksInPlane() { return this._displayOnlyPeaksInPlane; }
    get scene() { return this._scene; }
    get raycaster() { return this._raycaster; }

    // Move the light position so that it always stays with the camera.
    lightUpdate() {
        this._cameraLight.position.copy(this._scene.userData.camera.position);
    }

    /*
     * Render function called in animate.
     */
    render() {

        this._scene.userData.camera.updateMatrixWorld();
        if (displayMode.vr) {
            this._scene.userData.controls.update();
        }

        //controls.update();
        this._renderer.render(this._scene, this._scene.userData.camera);

        // Find the intersections.
        this._raycaster.setFromCamera(this._mouse, this._scene.userData.camera);
    }

    /*
     * On document mouse move event.
     */
    onDocumentMouseMove(event) {
        event.preventDefault();

        let mouseX = event.clientX - $("#" + this._container.id).offset().left;
        let mouseY = event.clientY - $("#" + this._container.id).offset().top;

        this._mouse.x = (mouseX / this._renderer.domElement.clientWidth) * 2 - 1;
        this._mouse.y = -(mouseY / this._renderer.domElement.clientHeight) * 2 + 1;
    }

    /*
     * On resize event.
     */
    onResize(width, height) {

        if (this._scene == undefined) {
            return;
        }

        this._scene.userData.camera.aspect = width / height;
        this._scene.userData.camera.updateProjectionMatrix();

        this._renderer.setSize(width, height);
    }
}

// Export the constructor.
module.exports = ReciprocalView3D;
