/**
 * 3D unit cell view.
 */

const THREE = require('three');
const OrbitControls = require('three-orbitcontrols');
const viewManager = require('../utils/view-manager');
const xtal = require('../../server/model/xtal/xtal');
const { sample } = require('../app-state');
const SeedRandom = require('seedrandom');
const atomColor = require('../utils/atom-color');

class UnitCellView3D {

    constructor() {
        
        this._renderer = null;
        this._width = 100;
        this._height = 100;

        let html = `
        <div style='padding-right:0;padding-left:0;width:100%;height:100%' id='unit3D'/>
        <div style='position: absolute; top: 0; left: 0; width: 0; height: 0;' id='layer3D'></div>`;

        // Using lambda function does not work here.
        myLayout.registerComponent('Unitcell View', function(container, state) {
            container.getElement().html(html);
        });

        // But it works here.
        myLayout.on('componentCreated', (component) => {

            // We can get the component main at its creation.
            if (component.componentName === 'Unitcell View') {

                // Store the initial size of the component.
                this._width = component.container.width;
                this._height = component.container.height;

                component.container.on('resize', () => {
                    this.onResize(component.container.width, component.container.height);
                });
            }
        });

        // Add the view to the view manager that is responsible of the animation.
        viewManager.addView(this);
    }

    init() {
        // Selects the div to place the three.js scene.
        this._container = document.getElementById('unit3D');

        // Antialias to true slows down the display but is required in the case of macOS.
        // Otherwise the display is flickering when there are multiple scissors.
        let antialias = false;
        if (process.platform === 'darwin') {
            antialias = true;
        }

        // Create the renderer.
        this._renderer = new THREE.WebGLRenderer({ antialias: antialias });

        // Set the initial size.
        this._renderer.setSize(this._width, this._height);
        this._container.appendChild(this._renderer.domElement);

        // basic info needed for unit cell
        this._spacegroup = sample.sg;
        this._atomnames = [];
        this._atompos = [];

        // init the graphics
        this._atomsGroup = new THREE.Group();
        this._ucGroup = new THREE.Group();
        this._textLabels = [];

        // Get the div containing the scene.
        this._panelDiv = document.getElementById('unit3D');
        this._layerDiv = document.getElementById('layer3D');

        this._scene = new THREE.Scene();
        this._scene.background = new THREE.Color(0xcccccc);

        const camera = new THREE.PerspectiveCamera(50, 1, 1, 10);
        camera.position.z = 2;
        this._scene.userData.camera = camera;

        const controls = new OrbitControls(this._scene.userData.camera, this._container);
        controls.minDistance = 2;
        controls.maxDistance = 5;
        controls.enablePan = false;
        controls.enableZoom = true;
        this._scene.userData.controls = controls;

        const light = new THREE.DirectionalLight(0xffffff, 0.5);
        light.position.set(1, 1, 1);
        this._scene.add(light);
        this._scene.add(new THREE.HemisphereLight(0xaaaaaa, 0x444444));

        // Add the view to the scene manager that is responsible of the animation.
        viewManager.addView(this);
        this.calcUC();


        // slot for updated unit cell infos
        PubSub.subscribe('redraw-unitcell', (message, values) => {
            for(let value in values) {
                if (value === "atomnames") {
                    this._atomnames = values[value];
                }
                else if (value === "atompos") {
                    this._atompos = values[value];
                }
                else if (value === "spacegroup") {
                    this._spacegroup = values[value];
                }
            }

            this.calcUC();
        });

        // Here ?
        PubSub.publish('update-atom-controller', 0);
    }

    get renderer() { return this._renderer; }

    render() {
        // Do not render text labels because it is consuming too much CPU.
        //this._updateTextLabels();
        this._renderer.render(this._scene, this._scene.userData.camera);
    }

    _updateTextLabels() {
        for(let idxChild=0; idxChild<this._layerDiv.children.length; ++idxChild) {
            const pos = this._layerDiv.children[idxChild].userData.pos;
            const name = this._layerDiv.children[idxChild].userData.name;

            const vecObj = new THREE.Vector3(pos[0], pos[1], pos[2]);
            vecObj.project(this._scene.userData.camera);
            const xView = this._renderer.domElement.clientLeft + 0.5 * (1. + vecObj.x) * this._renderer.domElement.clientWidth;
            const yView = this._renderer.domElement.clientTop + 0.5 * (1. - vecObj.y) * this._renderer.domElement.clientHeight;

            this._layerDiv.children[idxChild].style.position = "absolute";
            this._layerDiv.children[idxChild].style.left = xView + "px";
            this._layerDiv.children[idxChild].style.top = yView + "px";
            this._layerDiv.children[idxChild].innerHTML = name;
        }
    }


    clearUC() {
        // clear current atoms
        this._scene.remove(this._atomsGroup);
        for (let i = this._atomsGroup.children.length - 1; i >= 0; i--)
            this._atomsGroup.remove(this._atomsGroup.children[i]);

        // clear current unit cell
        this._scene.remove(this._ucGroup);
        for (let i = this._ucGroup.children.length - 1; i >= 0; i--)
            this._ucGroup.remove(this._ucGroup.children[i]);

        // clear text labels
        for(let i = 0; i < this._textLabels.length; ++i)
            this._layerDiv.removeChild(this._textLabels[i]);
        this._textLabels = [];
    }


    calcUC() {
        this.clearUC();

        // --------------------------------------------------------------------
        // unit cell
        const geoUC = new THREE.EdgesGeometry(new THREE.CubeGeometry(1,1,1));
        const matUC = new THREE.LineBasicMaterial({
          color: 0x000000, linewidth: 4,
        });
        const linesUC = new THREE.LineSegments(geoUC, matUC);
        this._ucGroup.add(linesUC);
        this._scene.add(this._ucGroup);
        // --------------------------------------------------------------------


        // --------------------------------------------------------------------
        // symmetrise atomic positions
        // don't do anything if the data is invalid
        if(this._atompos.length !== this._atomnames.length)
            return;

        const allpositions = xtal.getAllSymmetricPositionsWithinCell(this._spacegroup, this._atompos);
        const baseRadius = 0.04;

        for(let atomidx=0; atomidx<this._atomnames.length; ++atomidx) {
            const atomname = this._atomnames[atomidx];
            const allpos = allpositions[atomidx];

            const atomcolor = atomColor(atomname, this._atompos[atomidx][0]);

            for(let posidx=0; posidx<allpos.length; ++posidx) {
                const pos = allpos[posidx];

                // geometry
                const geometry = new THREE.SphereGeometry(baseRadius, 32, 32);
                const material = new THREE.MeshStandardMaterial({
                    color: atomcolor,
                    roughness: 0.5, metalness: 0, flatShading: true
                });

                const mesh = new THREE.Mesh(geometry, material);
                mesh.position.set(pos[0], pos[1], pos[2]);

                this._atomsGroup.add(mesh);


                // text label
                const div = document.createElement("div");
                div.userData = { "pos": pos, "name": atomname };
                this._layerDiv.appendChild(div);
                this._textLabels.push(div);
            }
        }

        this._scene.add(this._atomsGroup);
        //this._updateTextLabels();
        // --------------------------------------------------------------------
    }

    /*
     * On resize event.
     */
    onResize(width, height) {
        this._scene.userData.camera.aspect = width / height;
        this._scene.userData.camera.updateProjectionMatrix();

        this._renderer.setSize(width, height);
    }

    get scene() { return this._scene; }
}


// Export the constructor.
module.exports = UnitCellView3D;
