/**
 * Applies symmetry constraints to UI input elemets
 * @author Siqin Meng (siqin.meng@helmholtz-berlin.de)
 * @date Dec-2017
 */

const xtal = require('../xtal/xtal');

class SymmetryConstraint {
    /**
     * Constructor.
     */
    constructor() {
    };
    
    /**
     * Updates the input according to the space group.
     * @param {structure} inputs - the sample data.
     * @return {string[]} - Returns the string list of fixed values.
     */
    updateRules(inputs) {

        let fixedValues = [];

        let sg = inputs.sg;
        let currentRules = spaceGroupNumbers[findLatticeType(sg)].rules;

        for (let rule of currentRules) {
            let [left, right] = rule.split('=');
            
            fixedValues.push(left);

            if (/^[0-9]+$/.test(right)) {
                inputs[left] = right;
            } else {
                inputs[left] = inputs[right];
            }
        }
        
        return fixedValues;
    }
        
};

/**
* @param {number|string} sg - space group identifier, accepts either number or symbol.
* @returns {string} - Family of given space group.
*/
function findLatticeType(sg){
    let sg_no = 0;
    if (/^[0-9]+$/.test(sg)) { // number type converted automatically to string.
        sg_no = sg;
    } else {
        try {
            sg_no = xtal.spaceGroups[sg].no;
        } catch (err) {
            sg_no = 0;
        }
    }
    
    for (let lattice in spaceGroupNumbers) {
        if (sg_no >= spaceGroupNumbers[lattice].start && sg_no<= spaceGroupNumbers[lattice].end) {
            return lattice;
        }
    }
    return 'unknown';
}

const spaceGroupNumbers = {
    triclinic : {start: 1, end: 2, rules: []},
    monoclinic: {start: 3, end: 15, rules: ['aa=90', 'cc=90']},
    orthorhombic: {start: 16, end: 74, rules: ['aa=90', 'bb=90', 'cc=90']},
    tetragonal: {start: 75, end: 142, rules: ['bs=as', 'aa=90', 'bb=90', 'cc=90']},
    hexagonal: {start: 143, end: 194, rules: ['bs=as', 'aa=90', 'bb=90', 'cc=120']},
    cubic: {start: 195, end: 230, rules: ['bs=as', 'cs=as', 'aa=90', 'bb=90', 'cc=90']},
    unknown: {start:0, end: 0, rules: []}
}


module.exports = SymmetryConstraint;
