/**
 * S(Q,E) data from user-given formula
 * @author Tobias Weber <tweber@ill.fr>
 * @date Jan-2018
 * @license see "LICENSE" file
 */

const math = require("mathjs");
const sqwdata = require("./sqw-data");

/**
 * Normal distribution
 */
function gaussian(sigma, mu, x)
{
    const c = 1./(math.sqrt(2.*math.pi) * sigma);
    const a = 0.5 * (x-mu)/sigma * (x-mu)/sigma;
    return c * math.exp(-a);
}


class SqwFormula
{
    constructor(options = {})
    {
        this._sigmaE = 0.01;
        if(options.sigmaE != null)
            this._sigmaE = options.sigmaE;

        this._S0 = 1.;
        if(options.S0 != null)
            this._S0 = options.S0;

        this._maxS = 1.;
        if(options.maxS != null)
            this._maxS = options.maxS;

        this._minS = 0.;
        if(options.minS != null)
            this._minS = options.minS;

        const expr = options.formula;
        if(expr == null)
            return;

        this.loadExpr(expr);
    }


    loadExpr(expr)
    {
        try {
          this._expr = expr;
          this._ast = math.parse(expr);
          this._prog = this._ast.compile();
        }
        catch(err) {
            console.error(err);
            return false;
        }


        try {
            // the user can choose between
            // giving a dispersion relation "E = ..."
            // or the dynamical structure factor "S = ..."
            this._resultVar = this._ast.object.name;
        }
        catch(err) {
            // if nothing is given, assume S
            this._resultVar = "S";
        }
        
        return true;
    }


    query(h, k, l, E)
    {
        const vechklE = math.matrix([ h, k, l, E ]);
        const vecBragg = math.matrix([math.round(h), math.round(k), math.round(l), math.round(E)]);

        const vecq = math.subtract(vechklE, vecBragg);
        const q = math.norm(vecq);
        const Q = math.norm(vechklE);


        let result = 0.;

        try {
            // momentum variables
            let varmap = {
                "q":q, "qh":vecq.get([0]), "qk":vecq.get([1]), "ql":vecq.get([2]),
                "Q":Q, "h":h, "k":k, "l":l
            };

            // if S is queried, also add the energy
            if(this._resultVar === "S")
                varmap["E"] = E;

            result = this._prog.eval(varmap);
        }
        catch(err) {
            console.error(err);
        }


        let S = 0.;

        switch(this._resultVar) {
            case "E":
                const Emean = result;
                S = this._S0 * gaussian(this._sigmaE, Emean, E);

                // upper cutoff
                if(S > this._maxS)
                    S = this._maxS;
                // lower cutoff
                if(S < this._minS)
                    S = this._minS;
                break;

            case "S":
            default:
                S = result;
                break;
        }

        return { "S" : S, "coordinates" : vechklE.coordinates };
    }
};



// ----------------------------------------------------------------------------
/**
 * exports
 */
module.exports = {
    // class
    "Sqw" : SqwFormula,
};
// ----------------------------------------------------------------------------
