#!/usr/bin/env node --use_strict
/**
 * Standalone crystal calculation program
 * @author Tobias Weber <tweber@ill.fr>
 * @date 14-Nov-2017
 * @license see "LICENSE" file
 *
 * run with: node --use_strict run.js
 */

const xtal = require("./xtal");
const math = require("mathjs");
const readline = require("readline");

const io = readline.createInterface(
{
    input: process.stdin,
    output: process.stdout,
    prompt: "xtal> "
});


// ----------------------------------------------------------------------------

let g_spacegroup = "";
let g_atomnames = [];
let g_atompos = [];


function showHelp()
{
    console.log("Structure factor calculator, tweber@ill.fr, 2017\n");
    console.log("Commands:");
    console.log("    sg <name> ...................... set space group");
    console.log("    at <name> <x> <y> <z> .......... add atom");
    console.log("    cl ............................. clear data");
    console.log("    pr ............................. print data");
    console.log("    sf <h> <k> <l> ................. calculate structure factor");
    console.log("    sf <order> ..................... calculate list of structure factors");
    console.log("    sy ............................. calculate symmetric positions");
    console.log("    exit ........................... quit");
    console.log();
}


function setSpacegroup(sg)
{
    g_spacegroup = sg.valueOf();
}


function addAtom(name, x, y, z)
{
    g_atomnames.push(name);
    let pos = [ parseFloat(x), parseFloat(y), parseFloat(z) ];
    g_atompos.push(pos);
}


function printData()
{
    console.log("Space group: " + g_spacegroup);

    for(let idx=0; idx<g_atomnames.length; ++idx)
    {
        let name = g_atomnames[idx];
        let pos = g_atompos[idx];

        console.log("Atom " + (idx+1) + ": " + name + ", (" + pos + ")");
    }
}


function calcSymmetricPos()
{
    for(let idx=0; idx<g_atomnames.length; ++idx)
    {
        let name = g_atomnames[idx];
        let pos = g_atompos[idx];

        let sympos = xtal.getSymmetricPositions(g_spacegroup, pos);

        console.log("Atom " + (idx+1) + ": " + name);
        for(let ipos=0; ipos<sympos.length; ++ipos)
            console.log("\tPos. " + (ipos+1) + ": (" + sympos[ipos] + ")");
    }
}


function clearData()
{
    g_spacegroup = "";
    g_atomnames = [];
    g_atompos = [];
}


function calcSF(h, k, l)
{
    let latt = [ parseFloat(h), parseFloat(k), parseFloat(l) ];
    let absent = xtal.isPeakAbsent(latt, g_spacegroup);
    if(!absent)
    {
        console.log("Bragg peak is systematically allowed.");
        let sf = xtal.getStructureFactor(latt, g_atomnames, g_atompos, g_spacegroup);
        let intensity = math.multiply(sf, math.conj(sf));
        let sf_abs = math.sqrt(intensity);

        console.log("Structure factor: " + sf);
        console.log("Absolute structure factor: " + sf_abs);
        console.log("Intensity: " + intensity);
    }
    else
    {
        console.log("Bragg peak is systematically absent.");
    }
}


function calcSFList(maxorder)
{
    let order = parseInt(maxorder);

    console.log("h \tk \tl \tIntensity \tStructure factor");
    console.log("----------------------------------------------------------------------");

    let scatteringLengths = xtal.getAllScatteringLengths(g_atomnames);
    let sympos = xtal.getAllSymmetricPositions(g_spacegroup, g_atompos);

    for(let h=-order; h<=order; ++h)
        for(let k=-order; k<=order; ++k)
            for(let l=-order; l<=order; ++l)
            {
                const latt = [ h, k, l ];
                if(xtal.isPeakAbsent(latt, g_spacegroup))
                    continue;

                const sf = xtal.getStructureFactorDirect(latt, scatteringLengths, sympos);
                const [sf_str, sf2_str] = xtal.formatStructureFactor(sf, 8);

                console.log(latt[0] + "\t" + latt[1] + "\t" + latt[2] + "\t" +
                    sf2_str + "\t" + sf_str);
            }
    console.log("----------------------------------------------------------------------");
}


// ----------------------------------------------------------------------------


showHelp();
io.prompt();

io.on("line", (input) =>
{
    input = input.trim();
    let toks = input.split(" ");

    if(toks.length >= 1)
    {
        if(toks[0] === "?" || toks[0] === "help")
            showHelp();
        else if(toks[0] === "exit" || toks[0] === "quit")
            process.exit(0);
        else if(toks[0] === "sg" && toks.length >= 2)
            setSpacegroup(toks[1]);
        else if(toks[0] === "at" && toks.length >= 5)
            addAtom(toks[1], toks[2], toks[3], toks[4]);
        else if(toks[0] === "sf" && toks.length >= 4)
            calcSF(toks[1], toks[2], toks[3]);
        else if(toks[0] === "sf" && toks.length >= 2)
            calcSFList(toks[1]);
        else if(toks[0] === "pr")
            printData();
        else if(toks[0] === "sy")
            calcSymmetricPos();
        else if(toks[0] === "cl")
            clearData();
        else if (toks[0] === "")
            /*nop*/;
        else
            console.error("Unknown command.");
    }

    io.prompt();
});
